/*
 * evolve.cpp -- evolve an Iota source program.
 * 
 * To make sense of this file, see the video tutorial at:
 * http://vimeo.com/technotes/iota-genome
 *
 * Copyright 2012 David R. Miller
 *
 * The Iota simulator and this evolve.cpp source file are free software:
 * you can redistribute them and/or modify them under the terms of the
 * GNU General Public License as published by the Free Software Foundation,
 * version 3 of the License.
 *
 * Iota is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Iota and evolve.cpp. If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact: http://www.millermattson.com/dave
 */

#include <algorithm>
#include <cmath>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <iostream>
#include <sstream>
#include <string>
#include <thread>
#include <vector>
#include "iota-machine.h"

using namespace std;


// Compare string a with "Hello World!".
// Returns 0.0..1.0, where 1.0 is a perfect match.
//
float textSimilarity(const string &a)
{
    const string ref { "Hello World!" };
    float fitness = 0.0;

    for (size_t i = 0; i < ref.size() && i < a.size(); ++i) {
        fitness += (a[i] == ref[i]);
    }

    // Optional to evolve an exact-length solution:
    // Give credit for having the correct length;
    // Exact length counts as much as getting one char correct:
    //fitness += (a.size() == ref.size());
    //return fitness / (ref.size() + 1); // normalize to [0.0..1.0]

    return fitness / ref.size(); // normalize to [0.0..1.0]
}

// Compare string a with a reference string. Think of them as two abstract vectors.
// Calculate the root mean square difference between the two.
// Returns 0.0..1.0, where 1.0 is a perfect match.
// Assumes 8-bit chars.
//
float vectorSimilarityRMS(const string &a, const string &ref)
{
    size_t lengthToTest = min(a.size(), ref.size());

    if (lengthToTest == 0 || a.size() == 0 || ref.size() == 0) {
        return 0.0;
    }

    // Any missing elements count as maximum error
    float sumErrorsSquared = (float)(ref.size() - lengthToTest) * 256.0 * 256.0;

    auto itb = ref.begin();
    for (auto ita = a.cbegin(); ita != a.cend() && itb != ref.cend(); ++ita, ++itb) {
        long diff = (long)(unsigned char)*ita - (long)(unsigned char)*itb;
        sumErrorsSquared += (float)diff * (float)diff;
    }

    float rms = sqrt(sumErrorsSquared / (float)ref.size());
    return 1.0 / (1.0 + rms);
}


struct individual
{
    string mGenome;  // an Iota source program
    float mFitness;  // most recent fitness measure [0.0..1.0]
};

class universe
{
public:
    typedef unsigned short iotaArchType;    // underlying Iota architecture type
    typedef iota::machine<iotaArchType> iotaMachineType;

    universe(size_t archLength, size_t archWidth,
              size_t population, string &initialProgram);
    void measureFitness(size_t maxCyclesToRun, size_t numThreads = 1);
    void orderResults(void);
    void printBestResult(size_t maxCyclesToExecute);
    void reproduce(float fitnessDeletionThreshold, float eliteThreshold);
    void mutate(float mutationRatePerLetter);

private:
    void measureFitnessAsyncRange(iotaMachineType &myMachine,
            size_t beg, size_t end, size_t maxCyclesToRun);

    size_t mArchLength;              // Iota architecture number of memory locations
    size_t mArchWidth;               // Iota architecture width of memory in bits
    vector<individual> mPopulation;  // our individuals to evolve
    vector<iotaMachineType> mIotaMachines;   // a pool of Iota virtual machines to use
    vector<thread> mThreadPool;      // for multiple threads during fitness testing
};

universe::universe(size_t archLength, size_t archWidth,
                     size_t population, string &initialProgram)
    : mArchLength(archLength), mArchWidth(archWidth),
      mPopulation(vector<individual>(population, individual()))
{
    // At this point, all individuals have a 0.0 fitness and an
    // empty mGenome string.

    // Copy a starting genome to all individuals:
    for (auto &ind : mPopulation) {
        ind.mGenome = initialProgram;
    }
}

void universe::orderResults()
{
    sort(mPopulation.begin(), mPopulation.end(),
        [](const individual &a, const individual &b)
            {
                return a.mFitness > b.mFitness;
            });

    // If we evolved a perfect solution, report it now:

    if (mPopulation[0].mFitness >= 1.0) {
        cout << "Solved!" << endl;
        cout << "Genome = \'" << mPopulation[0].mGenome << "\'" << endl;

        // Tip: the final genome can be run through the iota command line
        // interpreter to see how the genome works to produce the intended output.

        // We're done here
        exit(0);
    }
}

// Show the best fitness individual -- show its genome, and
// re-run its genome through the Iota simulator to get its output to show.
//
void universe::printBestResult(size_t maxCyclesToExecute)
{
    auto &ind = mPopulation[0];

    // The Iota machine output strings are not saved with the individuals, so to see
    // the output, we need to re-run the Iota program:

    // Grab any ol' virtual machine to use temporarily:
    auto &myMachine = mIotaMachines[0]; // borrow the first one
    myMachine.reset();
    myMachine.copyProgramToMachineMemory(ind.mGenome);
    myMachine.executeAtMost(maxCyclesToExecute);

    auto &output = myMachine.mOutputStream;
    cout << "Best output (len=" << output.size() << ") = \"" << output << "\"" << endl;
    cout << "Best genome (" << ind.mFitness << ") = \'" << ind.mGenome << "\'" << endl;
}


void universe::measureFitness(size_t maxCyclesToRun, size_t numThreads)
{
    // Clear out our container of threads (reuse the container)
    mThreadPool.clear();

    // Divide the population among numThreads threads. (This is approximate).

    size_t numElementsPerThread = max(1UL, 1 + mPopulation.size() / numThreads);
    size_t first = 0;
    size_t end = min(mPopulation.size(), first + numElementsPerThread);

    // Make sure our pool of Iota simulators has enough simulators for all the threads.
    // This design assumes that there are few threads that all run in parallel, and
    // each queued thread gets a dedicated simulator machine for its exclusive use.

    while (mIotaMachines.size() < numThreads) {
        mIotaMachines.push_back(iotaMachineType(mArchLength, mArchWidth));
    }

    // Launch the threads

    size_t threadNum = 0;
    while (first < mPopulation.size() && end <= mPopulation.size()) {
        // cout << "Thread first,last = " << first << " to " << end << endl;

        mThreadPool.push_back(
                thread(&universe::measureFitnessAsyncRange,
                        this, ref(mIotaMachines[threadNum]),
                        first, end, maxCyclesToRun));

        first += numElementsPerThread;
        end = min(mPopulation.size(), first + numElementsPerThread);
        ++threadNum;
    }

    // Join -- wait for all the threads to finish

    for (auto &th : mThreadPool) {
        th.join();
    }
}

// Returns [0.0..1.0] where 1.0 means a perfect solution.
// This function reads and writes only those individuals in its
// prescribed range and doesn't modify the container, so there is no
// synchronization needed with other threads that do the same.
//
void universe::measureFitnessAsyncRange(iotaMachineType &myMachine,
            size_t beg, size_t end, size_t maxCyclesToRun)
{
    for (size_t idx = beg; idx != end; ++idx) {
        auto &ind = mPopulation[idx];

        // Reset the Iota VM and load the program:
        myMachine.reset();
        myMachine.copyProgramToMachineMemory(ind.mGenome);

        // Execute the program and test the output:
        myMachine.executeAtMost(maxCyclesToRun);
        ind.mFitness = textSimilarity(myMachine.mOutputStream); // test for "Hello World!"
        //ind.mFitness = vectorSimilarityRMS(myMachine.mOutputStream, string { "Hello World!" }); // test for "Hello World!"
    }

    // At this point, all the individuals' fitnesses have been updated,
    // for all the individuals in this range.
}


// Quick and easy who-mates-and-who-dies algorithm using two thresholds.
// The thresholds are between 0.0 and 1.0, and are percentiles for the
// current set of fitness measurements after sorting.
// The diagram below shows two typical threshold values which can be
// changed with command line options -F and -E.
//
// +--------------------------------------------+ 1.0 best fitness
// | Always a parent; lives another generation  |
// +--------------------------------------------+ ~0.9 <= Elite threshold (-E)
// |                                            |
// | May be a parent; lives another generation  |
// |                                            |
// +--------------------------------------------+ ~0.5 <= Fitness Deletion Threshold (-F)
// |                                            |
// | Does not survive the generation,           |
// | Replaced by new offspring                  |
// |                                            |
// +--------------------------------------------+ 0.0 worst fitness
//
// This function iterates through individuals from the worst fitness to the
// fitness deletion threshold, replacing each with a new individual. Rather
// than deleting an object only to have to create a replacement, we just
// simply overwrite the genome with a new one derived from two parents
// selected from above the thresholds. A new genome makes a new individual.
//
void universe::reproduce(float fitnessDeletionThreshold, float eliteThreshold)
{
    size_t fitnessThreshIdx = mPopulation.size() * (1.0 - fitnessDeletionThreshold);
    size_t eliteThreshIdx   = mPopulation.size() * (1.0 - eliteThreshold);

    for (size_t replaceIdx = fitnessThreshIdx + 1; replaceIdx < mPopulation.size(); ++replaceIdx) {
        // Select two parents, one from above the elite threshold, the other will be
        // a random selection above the fitness deletion threshold, which might or might
        // not be from the elite section.

        size_t parent0Idx = rand() % (eliteThreshIdx + 1);
        size_t parent1Idx = rand() % (fitnessThreshIdx + 1);

        // Favor higher fitness parents by finding another set of random
        // indices between 0 and the indices just calculated:

        parent0Idx = rand() % (parent0Idx + 1);
        parent1Idx = rand() % (parent1Idx + 1);

        individual &parent0 = mPopulation[parent0Idx];
        individual &parent1 = mPopulation[parent1Idx];

        // Pick some random splice points:
        // p0 and p1 will be indices, ordered, both in the range of parent0 and parent1:

        size_t minLength = min(parent0.mGenome.size(), parent1.mGenome.size());
        size_t p0 = rand() % (minLength + 1);
        size_t p1 = rand() % (minLength + 1);
        if (p0 > p1) {
            swap(p0, p1);
        }

        // Here are some recombination strategies to use as a starting point.
        // Experiment!

        string newGenome;
        switch (rand() % 3) {
        case 0:        // AaaaabbbbbaaaaaAaaaa
            newGenome =
                    parent0.mGenome.substr(0, p0)
                  + parent1.mGenome.substr(p0, p1 - p0)
                  + parent0.mGenome.substr(p1)
                  + parent0.mGenome.substr(0, p0);
            break;

        case 1:        // Aaaaabbbbbaaaaa
            newGenome =
                    parent0.mGenome.substr(0, p0)
                  + parent1.mGenome.substr(p0)
                  + parent0.mGenome.substr(p1);
            break;

        default:
        case 2:        // aaaaabbbbbAaaaa
            newGenome =
                    parent0.mGenome.substr(p0, p1 - p0)
                  + parent1.mGenome.substr(p1)
                  + parent0.mGenome.substr(p0);
            break;
        }

        // Trim the new genome to the Iota memory length:
        newGenome = newGenome.substr(0, mArchLength);

        // Make a new individual:
        mPopulation[replaceIdx].mGenome = newGenome;
        mPopulation[replaceIdx].mFitness = 0.0;
    }
}

// This mutates some individual's genomes by randomly changing
// a single character to a random Iota opcode letter.
//
void universe::mutate(float mutationRatePerLetter)
{
    //const string &programCharset = mIotaMachines[0].allOpcodeChars;

    for (auto &ind : mPopulation) {
        if (rand() / (float)RAND_MAX < mutationRatePerLetter * ind.mGenome.size()) {
            size_t numCharsToMutate =
                    (int)(mutationRatePerLetter * ind.mGenome.size()
                            + 0.5 + ((rand() / (float)RAND_MAX)));
            for (auto i = numCharsToMutate; i > 0; --i) {
                size_t idx = rand() % ind.mGenome.size();
                // This replaces a char with a random char from the Iota opcode set:
                //ind.mGenome[idx] = programCharset[rand() % programCharset.size()];

                // or, alternatively, the following line replaces a char with any
                // random value that fits in the logical Iota memory width:
                //ind.mGenome[idx] = rand() % (1UL << mArchWidth);

                // or, alternatively, the following line replaces a char with any
                // random *printable* char for a prettier display during evolution:
                ind.mGenome[idx] = 32 + (rand() % (127 - 32));
            }
        }
    }
}


// Above this line is all about defining class universe
// ***********************************************************************************************************************
// Below this line is all about using class universe


// Command line args and their defaults. This encapsulates all the
// command line options, their types, and their default values in
// one easy-to-edit format and a simple interface. In main(), just
// instantiate one of these, then call its parseCommandLine() member.
//
// This quick-n-dirty parser implementation does not validate option
// arguments and does not report multiple occurrences of the same option.
// Replace this with a commercial quality parser if that's what you need.
//
struct commandLineParameters
{
    commandLineParameters(void) :
        archLength(32),        // -L
        archWidth(8),          // -W
        fitnessThreshold(0.5), // -F
        eliteThreshold(0.9),   // -E
        maxCyclesToRun(200),   // -k
        maxGenerations(1e9),   // -g
        mutationRate(0.01),    // -M
        numThreads(1),         // -t
        population(100),       // -p
        program(";H")          // the only arg with no option letter
    {}

    size_t archLength;
    size_t archWidth;
    float fitnessThreshold;
    float eliteThreshold;
    size_t maxCyclesToRun;
    size_t maxGenerations;
    float mutationRate;
    size_t numThreads;
    size_t population;
    string program;

    void parseCommandLine(int argc, char **argv);

private:
    template <typename T>
    void getArg(T *pRetVal, int &i, int argc, char **argv)
    {
        if (argv[i][2] != '\0') {
            istringstream(&argv[i][2]) >> *pRetVal;
            return;
        } else if (i < argc - 1) {
            istringstream(argv[++i]) >> *pRetVal;
            return;
        }

        cout << "synopsis: evolve [-p population] [-g maxGenerations]" << endl;
        cout << "[-k maxCyclesToRun] [-t numThreads] [-L archLength] [-W archWidth]" << endl;
        cout << "[-M mutationRate] [-F fitnessThreshold] [-E eliteThreshold] [startingProgram]" << endl;
        exit(1);  // alternatively, return a status to the caller
    }
};

// Sets all the values in the parameter block, or else issues a synopsis
// message and exits. If the same option appears multiple times, only the
// last one matters.
//
void commandLineParameters::parseCommandLine(int argc, char **argv)
{
    for (int i = 1; i < argc; ++i) {
        if (strncmp(argv[i], "-M", 2) == 0) {
            getArg(&mutationRate, i, argc, argv);
        } else if (strncmp(argv[i], "-p", 2) == 0) {
            getArg(&population, i, argc, argv);
        } else if (strncmp(argv[i], "-g", 2) == 0) {
            getArg(&maxGenerations, i, argc, argv);
        } else if (strncmp(argv[i], "-k", 2) == 0) {
            getArg(&maxCyclesToRun, i, argc, argv);
        } else if (strncmp(argv[i], "-t", 2) == 0) {
            getArg(&numThreads, i, argc, argv);
        } else if (strncmp(argv[i], "-L", 2) == 0) {
            getArg(&archLength, i, argc, argv);
        } else if (strncmp(argv[i], "-W", 2) == 0) {
            getArg(&archWidth, i, argc, argv);
        } else if (strncmp(argv[i], "-F", 2) == 0) {
            getArg(&fitnessThreshold, i, argc, argv);
        } else if (strncmp(argv[i], "-E", 2) == 0) {
            getArg(&eliteThreshold, i, argc, argv);
        } else {
            program = string(argv[i]);
        }
    }
}


int main(int argc, char **argv)
{
    // Disable the next line to debug with a predictable random sequence:
    srand(time(NULL));

    commandLineParameters args;
    args.parseCommandLine(argc, argv);

    // Make a universe with a bunch of individuals:
    universe myWorld(args.archLength, args.archWidth, args.population, args.program);

    // Evolve the population
    const size_t everyNth = 32;
    for (size_t generationNumber = 0; generationNumber < args.maxGenerations; ++generationNumber) {
        if (generationNumber % everyNth == 0) {
            cout << "\nGeneration " << generationNumber << endl;
        }

        myWorld.measureFitness(args.maxCyclesToRun, args.numThreads);
        myWorld.orderResults();
        if (generationNumber % everyNth == 0) {
            myWorld.printBestResult(args.maxCyclesToRun);
        }
        myWorld.reproduce(args.fitnessThreshold, args.eliteThreshold);
        myWorld.mutate(args.mutationRate);
    }
}

