/*
 * iota-machine.h - interface for the Iota little language simulator
 *
 * Copyright 2012 David R. Miller
 *
 * This file is part of the Iota little language simulator.
 *
 * Iota is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Iota is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Iota.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact: http://www.millermattson.com/dave
 */

/**
 * @file iota-machine.h
 * Interface for the Iota little language simulator
 *
 * @author David R. Miller, dave@millermattson.com
 * @version 1.0
 * @date October, 2012
 *
 * Compiles with gcc 4.7 with options:
 * -std=c++11 -W -Wall -Wextra -pedantic -D__GXX_EXPERIMENTAL_CXX0X__=1
 */

#ifndef IOTA_MACHINE_H
#define IOTA_MACHINE_H

#include <cstdlib>
#include <ctime>
#include <iomanip>
#include <iostream>
#include <stdexcept>
#include <string>
#include <vector>


/**
 * @namespace iota
 * The iota namespace encapsulates the Iota simulator interface.
 */
namespace iota {

    static const std::string languageVersion { "Iota language version 1.0" };
    static const std::string simulatorVersion { "Iota simulator version 1.0" };

    /**
     * @class iotaExcept
     * This is the object thrown by iota::machine() for all construction errors.
     */
    class iotaExcept : public std::runtime_error {
    public:
        iotaExcept() : std::runtime_error("Can't construct an iota::machine object (check the ctor args)") { }
    };

    /**
     * @class machine
     * An instance of this class simulates an Iota processing machine.
     * @tparam T The unsigned type that will be the underlying representation
     * for Iota memory locations. It must be an unsigned type at least
     * @c archWidth bits wide.
     */
    template <typename T>
    class machine
    {
    public:  // functions
        /// Exposes the template param, an unsigned integral type.
        typedef T value_type;

        /// Construct a virtual machine with memory archLength long and archWidth wide
        machine(size_t archLength, size_t archWidth = 8);

        /// Reset the virtual machine: PC = SP = 0; NZ = false, clear output string
        void reset(void);

        /// Given a source string of single-char mnemonics, copy to Iota memory.
        void copyProgramToMachineMemory(const std::string &src);

        /// Define a string of chars to supply the IN opcode.
        void setDataInputSource(const std::string &dataInSource);

        /// Zero for no debug output; nonzero gives additional output.
        void setDebugLevel(int level);

        /// Any nonzero value enables Iota execution trace to stdout
        void setExecutionTrace(int level);

        /// List the Iota program current in memory. size 0 means archLength.
        std::string listProgramShortMnemonics(size_t maxLength = 0) const;

        /// Print memory and register contents.
        void sprintState(void) const;

        /// Generate a listing with long mnemonics of the Iota program in memory.
        void disassemble(void) const;

        /// Execute the Iota program until it encounters HALT or executes this many cycles.
        size_t executeAtMost(size_t numberOfCycles);

        /// Execute the next Iota instruction.
        bool executeOneCycle(void);

    public: // directly accessible data
        /**
         * The container that collects the characters output by the Iota
         * @c OUT opcode. (An alternative implementation could have been
         * made using an @c ostringstream or @c ostream reference, where the
         * caller owns the container.) The size of the string is capped
         * at @c machine::maxOutputStreamSize characters.
         */
        std::string mOutputStream;     // destination for OUT opcode

        /// Some reasonably big limit for @c mOutputStream
        const size_t maxOutputStreamSize = 128*1024; // some reasonably big limit

        /**
         * At runtime, a list of all the single-character Iota opcode
         * mnemonics is collected in this container, derived from the
         * master opcode table @c opcodeTable.
         * @see opcodeTable.
         */
        std::string allOpcodeChars;   // a generated set of all defined short mnemonics

        /**
         * The master list of Iota opcode numbers. Determines the opcode numbers only.
         * @see findByNumber() and findByShortMnemonic()
         */
        enum opcodeNumber {
            OP_NOP,     // (0) No-operation, do-nothing
            OP_RESET,   // Reset the virtual machine
            OP_HALT,    // Halt Iota program execution
            OP_IN,      // Input a char from stdin, push it onto the stack
            OP_OUT,     // Pop a word from the stack, output to stdout
            OP_POP,     // Pop a word from the stack
            OP_DUP,     // Duplicate the last stacked value
            OP_PUSHPC,  // Push the PC onto the stack
            OP_POPPC,   // Pop the PC from the stack
            OP_POPSP,   // Pop the SP from the stack
            OP_SPTGT,   // (10) Set the SP to the next TARGET opcode
            OP_PUSHNZ,  // Push the NZ flag
            OP_SWAP,    // Swap the top two items on the stack
            OP_PUSH0,   // Push a zero onto the stack
            OP_ADD,     // Add the top two stacked words, push the result
            OP_SUB,     // Subtract the top two stacked words and push the result
            OP_INC,     // Increment the item at the top of the stack
            OP_DEC,     // Decrement the item on the top of the stack
            OP_MUL,     // Multiply the top two stacked words and push the result
            OP_DIV,     // Pop two words, divide, push the quotient and remainder
            OP_XOR,     // (20) Bitwise XOR the top two stacked words and push the result
            OP_AND,     // Bitwise AND the top two stacked words and push the result
            OP_OR,      // Bitwise OR the top two stacked words and push the result
            OP_SHL,     // Logical shift left the top stacked word
            OP_SHR,     // Logical shift right the top stacked word
            OP_NOT,     // Bitwise NOT the top stacked word
            OP_BZ,      // Branch if zero (NZ flag is false)
            OP_BNZ,     // Branch if nonzero (NZ flag is true)
            OP_BEQ,     // Compare top two stacked words, branch if equal
            OP_BGT,     // Compare top two stacked words, branch if greater than
            OP_BLT,     // (30) Compare top two stacked words, branch if less than
            OP_BGE,     // Compare top two stacked words, branch if greater than or equal
            OP_LOOP,    // Repeat the following instructions up to the next std::endl
            OP_ENDL,    // End of LOOP
            OP_BRAN,    // Branch to the next TARGET opcode
            OP_BRAP,    // Branch to the previous TARGET opcode
            OP_TARGET,  // Branch target for BRAN and BRAP
            OP_SKIP1,   // Skip one instruction
            OP_SKIP2,   // Skip the next two instructions
            OP_SKIP3,   // Skip the next three instructions
            OP_SKIP4,   // (40) Skip the next four instructions
            OP_SKIP5,   // Skip the next five instructions
            OP_SKIP6,   // Skip the next six instructions
            OP_SKIP7,   // Skip the next seven instructions
            OP_SKIP8,   // Skip the next eight instructions
            OP_SKIP9,   // (45) Skip the next nine instructions
            NUMBER_OF_OPCODES   // (46)
        };

        // *************************** under the covers ***************************

    private: // private types and data
        size_t mArchLength;             // simulated memory length [1..2^32)
        size_t mArchWidth;              // simulated memory width in bits [6..32)
        unsigned long mCycleNumber;     // number of Iota cycles executed
        size_t pc;                      // program counter
        size_t sp;                      // stack pointer
        bool   nz;                      // non-zero flag; true means non-zero
        std::string mDataIn;            // source of input data for IN opcode
        size_t mDataInIdx;              // index into dataIn
        int mExecutionTraceLevel;       // 0=no trace listing; >0 = trace listing
        int mDebugLevel;                // zero is silent; nonzero is extra debug output
        std::vector<T> mem;             // Iota simulated memory, archLength x archWidth

        struct opcodeDetails
        {
            enum opcodeNumber num;      // the opcode number, [0..NUMBER_OF_OPCODES)
            char shortMnemonic;         // the short opcode mnemonic, e.g, 'H'
            const char *longMnemonic;   // the long opcode mnemonic, e.g., "HALT"
        };

        // For performance reasons, it's important to have a constant-time
        // lookup tables by number and by char value. Since the tables are
        // so small and simple, we'll just build the two lookup tables at
        // runtime using the data in mOpcodeTable[] as input.

        static const opcodeDetails mOpcodeTable[NUMBER_OF_OPCODES];   // input, human-maintainable
        opcodeDetails mOpcodeTableByNumber[NUMBER_OF_OPCODES];        // generated at runtime
        opcodeDetails mOpcodeTableByCharacter[256];                   // generated at runtime

    private: // private functions
        // Wrap an address to the actual number of memory locations:
        inline T wrap(T a) { return a % mArchLength; }

        // Truncate a number to the machine word width:
        inline T trunc(T val) { return val & ((1UL << mArchWidth) - 1); }

        // Helpers for PC and SP arithmetic:
        inline void setnz(T val) { nz = !!val; }
        inline void decsp(void) { sp = sp == 0 ? mArchLength - 1 : sp - 1; }
        inline void incsp(void) { sp = (sp + 1) % mArchLength; }
        inline void incpc(void) { pc = (pc + 1) % mArchLength; }

        // create opcodeTableByNumber and opcodeTableByCharacter from mOpcodeTable
        void initOpcodeTables(void);

        // Given an opcode number, return its details. If the number is
        // not a valid opcode number, return the details for the NOP opcode.
        const opcodeDetails& findByNumber(size_t n) const;

        // Given an opcode single-character mnemonic. return its
        // details. If the char is not a valid opcode mnemonic, return
        // the details as if it were a NOP opcode.
        const opcodeDetails& findByShortMnemonic(char c) const;
    };


    // ****************************** Implementation ******************************


    /**
     * The master opcode table.
     * This list can be in any order. One or more sorted lists
     * will be derived at runtime from this table. This is what
     * maps the opcode numbers to mnemonics.
     */
    template <typename T>
    const typename machine<T>::opcodeDetails machine<T>::mOpcodeTable[NUMBER_OF_OPCODES] = {
        { OP_NOP,    ';', "NOP" },
        { OP_RESET,  'R', "RESET" },
        { OP_HALT,   'H', "HALT" },
        { OP_IN,     'I', "IN" },
        { OP_OUT,    'O', "OUT" },
        { OP_POP,    'p', "POP" },
        { OP_DUP,    'D', "DUP" },
        { OP_PUSHPC, 'C', "PUSHPC" },
        { OP_POPPC,  'c', "POPPC" },
        { OP_POPSP,  'Y', "POPSP" },
        { OP_SPTGT,  'G', "SPTGT" },
        { OP_PUSHNZ, 'P', "PUSHNZ" },
        { OP_SWAP,   'S', "SWAP" },
        { OP_PUSH0,  '0', "PUSH0" },
        { OP_ADD,    '+', "ADD" },
        { OP_SUB,    '-', "SUB" },
        { OP_INC,    '.', "INC" },
        { OP_DEC,    ',', "DEC" },
        { OP_MUL,    '*', "MUL" },
        { OP_DIV,    '/', "DIV" },
        { OP_XOR,    '^', "XOR" },
        { OP_AND,    '&', "AND" },
        { OP_OR,     '|', "OR" },
        { OP_SHL,    '(', "SHL" },
        { OP_SHR,    ')', "SHR" },
        { OP_NOT,    '~', "NOT" },
        { OP_BZ,     'Z', "BZ" },
        { OP_BNZ,    'z', "BNZ" },
        { OP_BEQ,    '=', "BEQ" },
        { OP_BGT,    '>', "BGT" },
        { OP_BLT,    '{', "BLT" },
        { OP_BGE,    '}', "BGE" },
        { OP_LOOP,   'L', "LOOP" },
        { OP_ENDL,   ']', "ENDL" },
        { OP_BRAN,   'B', "BRAN" },
        { OP_BRAP,   'b', "BRAP" },
        { OP_TARGET, 'T', "TARGET" },
        { OP_SKIP1,  '1', "SKIP1" },
        { OP_SKIP2,  '2', "SKIP2" },
        { OP_SKIP3,  '3', "SKIP3" },
        { OP_SKIP4,  '4', "SKIP4" },
        { OP_SKIP5,  '5', "SKIP5" },
        { OP_SKIP6,  '6', "SKIP6" },
        { OP_SKIP7,  '7', "SKIP7" },
        { OP_SKIP8,  '8', "SKIP8" },
        { OP_SKIP9,  '9', "SKIP9" }
    };

    /**
     * Constructs an Iota virtual machine.
     * @param[in] archWidth The width in bits of the Iota memory system, in
     * the range [6..32).
     * @param[in] archLength The number of memory locations in the Iota
     * virtual machine, in the range [1..2^32).
     * @exception iotaExcept Throws @c iota::iotaExcept if any ctor argument is
     * out of range.
     */
    template <typename T>
    machine<T>::machine(size_t archLength, size_t archWidth)
        : mOutputStream(std::string()), mArchLength(archLength), mArchWidth(archWidth),
          mCycleNumber(0), pc(0), sp(0), nz(0), mDataIn(std::string()), mDataInIdx(0),
          mExecutionTraceLevel(0), mDebugLevel(0), mem(std::vector<T>(archLength, 0))
    {
        // check ranges
        if (mArchWidth < 6 || mArchWidth > 32) {
            std::cerr << "Oops: archWidth must be in the range [6..32]" << std::endl;
            throw iotaExcept();
            exit(1);
        }

        if (mArchLength < 1) {
            std::cerr << "Oops: archLength must be in the range [1..2^32]" << std::endl;
            throw iotaExcept();
            exit(1);
        }

        initOpcodeTables();
    }

    /**
     * Reset the Iota virtual machine. This sets the @c PC and @c SP to location
     * zero, clears the @c NZ flag, resets the execution cycle counter
     * to zero, clears the output stream, and clears simulated memory to all
     * zero (OP_NOP opcodes)
     */
    template <typename T>
    void machine<T>::reset(void)
    {
        pc = 0;
        sp = 0;
        nz = 0;
        mCycleNumber = 0;
        mOutputStream = std::string();
        mDataInIdx = 0;
        fill(begin(mem), end(mem), OP_NOP);
    }

    /**
     * Given a string of characters, regard it as the source for an Iota
     * program and copy it to the Iota simulated memory starting at location
     * zero. Characters that are valid opcodes are stored in Iota memory as
     * the opcode number, otherwise the char value is cast to type @c T and
     * stored.
     * @param[in] src An Iota source program, one character per opcode.
     * If the string is longer than Iota memory, the remainder of the string
     * that doesn't fit is ignored.
     */
    template <typename T>
    void machine<T>::copyProgramToMachineMemory(const std::string &src)
    {
        if (src.size() == 0) {
            mem[0] = OP_NOP;
            return;
        }

        size_t loc = 0;
        for (auto it = src.begin(); loc < mArchLength && it != src.end(); ++loc, ++it) {
            // get and store the opcode number or the literal char value if not an opcode
            mem[loc] = (T)findByShortMnemonic(*it).num;
        }
    }

    /**
     * This specifies a string (copied into the Iota machine) that will
     * supply the stream of characters to be consumed by the Iota
     * @c IN opcode, one character from @c dataInSource for each @c IN
     * instruction. When the input reaches the end of the string, subsequent
     * @c IN instructions will repeatedly retrieve the last character of the
     * input string as though the input was an infinitely long string.
     * @param[in] dataInSource A string that will supply characters for the Iota
     * @c IN instruction.
     */
    template <typename T>
    void machine<T>::setDataInputSource(const std::string &dataInSource)
    {
        mDataIn = dataInSource;
        mDataInIdx = 0;
    }

    /**
     * Setting a nonzero debug level will cause extra output to be sent to
     * std::cerr (stderr).
     * @param[in] level Zero for no debug output; nonzero for more.
     */
    template <typename T>
    void machine<T>::setDebugLevel(int level)
    {
        mDebugLevel = level;
    }

    /**
     * A nonzero level causes a line to be printed to std::cerr (stderr) for
     * each Iota opcode executed.
     * @param[in] level Zero for no trace; nonzero to enable execution trace.
     */
    template <typename T>
    void machine<T>::setExecutionTrace(int level)
    {
        mExecutionTraceLevel = level;
    }

    /**
     * Generate a listing of the Iota program currently in memory. Memory
     * locations that contain a valid opcode number are represented by its
     * single-character opcode mnemonic, otherwise the value in memory is
     * cast to a @c char and represented by a single character.
     * @param[in] maxLength If zero (the default), the entire contents of Iota memory
     * will be formatted in source form, otherwise @c maxLength specifies the number
     * of memory locations to list in source form, starting at location zero.
     * @return The generated Iota source code, by value.
     */
    template <typename T>
    std::string machine<T>::listProgramShortMnemonics(size_t maxLength) const
    {
        std::string program { "" };

        if (maxLength == 0) {
            maxLength = mArchLength;
        }

        for (size_t i = 0; i < maxLength; ++i) {
            size_t n = (size_t)mem[i];
            if (n < NUMBER_OF_OPCODES) {
                program += findByNumber(mem[i]).shortMnemonic;
            } else if (n <= 0xff) {
                // show the literal character
                program += (char)n;
            } else {
                program += findByNumber(OP_NOP).shortMnemonic; // i.e., ';'
            }
        }

        return program;
    }

    /**
     * Print to std::cerr (stderr) the state of Iota memory and registers,
     * and a one-line disassembly of the next Iota opcode to be executed.
     */
    template <typename T>
    void machine<T>::sprintState(void) const
    {
        unsigned hexWidth = (mArchWidth - 1) / 4 + 1;

        std::cerr << "T"
             << std::setw(8) << std::dec << std::setfill('0') << mCycleNumber << ": "
             << "PC=" << std::setw(hexWidth) << std::hex << std::setfill('0') << pc << " "
             << "SP=" << std::setw(hexWidth) << std::hex << std::setfill('0') << sp << " "
             << "NZ=" << (nz ? 'T' : 'F') << " mem[]=";

        for (auto i : mem) {
            std::cerr << std::setw(hexWidth) << std::hex << std::setfill('0') << (size_t)i << " ";
        }

        std::cerr << std::endl
             << std::setw(hexWidth) << std::hex << std::setfill('0') << (size_t)pc << ": "
             << std::setw(hexWidth) << std::hex << std::setfill('0') << (size_t)mem[pc] << " "
             << findByNumber(mem[pc]).shortMnemonic << " "
             << findByNumber(mem[pc]).longMnemonic << " "
             << std::endl;
    }

    /**
     * Print to std::cerr (stderr) a disassembly, using long opcode mnemonics,
     * of the Iota simulated memory and registers.
     */
    template <typename T>
    void machine<T>::disassemble(void) const
    {
        std::cerr << "Memory dump/disasm:" << std::endl;

        for (size_t i = 0; i < mem.size(); ++i) {
            std::cerr << std::setw(3) << std::hex << std::setfill(' ') << i << " "
                 << std::setw(2) << std::hex << std::setfill('0') << (size_t)mem[i] << " ";
            size_t n = mem[i];
            if (n < NUMBER_OF_OPCODES) {
                std::cerr << findByNumber((size_t)n).shortMnemonic << " "
                     << findByNumber((size_t)n).longMnemonic;
            } else if (n <= 0xff) {
                std::cerr << "  \'" << (char)n << "\' (NOP)";
            } else {
                std::cerr << std::hex << n;
            }

            std::cerr << std::endl;
        }
    }

    /**
     * A wrapper around @c executeOneCycle().
     * @param[in] numberOfCycles The maximum number of times to call
     * @c executeOneCycle(). The Iota program will terminate prior to that
     * if it executes a HALT instruction.
     * @return The actual number of cycles executed.
     * @see @c executeOneCycle()
     */
    template <typename T>
    size_t machine<T>::executeAtMost(size_t numberOfCycles)
    {
        if (mExecutionTraceLevel) {
            sprintState();
        }

        size_t i;
        for (i = 0; i < numberOfCycles; ++i) {
            if (!executeOneCycle()) {
                return i;
            } else {
                if (mExecutionTraceLevel) {
                    sprintState();
                }
            }
        }

        return i;
    }

    // create opcodeTableByNumber and mOpcodeTableByCharacter from opcodeTable
    // create allOpcodeChars = a set of all defined short mnemonics
    //
    template <typename T>
    void machine<T>::initOpcodeTables(void)
    {
        // First, initialize opcodeTableByCharacter so that the num field
        // equals the char code, then replace the ones that are real opcodes:

        for (size_t i = 0; i < sizeof(mOpcodeTableByCharacter) / sizeof(mOpcodeTableByCharacter[0]); ++i) {
            mOpcodeTableByCharacter[i].longMnemonic = nullptr;
            mOpcodeTableByCharacter[i].shortMnemonic = (char)i;
            mOpcodeTableByCharacter[i].num = (decltype(mOpcodeTableByCharacter[i].num))(char)i;
        }

        for (auto details : mOpcodeTable) {
            mOpcodeTableByNumber[details.num] = details;
            mOpcodeTableByCharacter[(unsigned)details.shortMnemonic] = details;
            allOpcodeChars += details.shortMnemonic;
        }
    }

    template <typename T>
    const typename machine<T>::opcodeDetails& machine<T>::findByNumber(size_t n) const
    {
        if (n < NUMBER_OF_OPCODES) {
            return mOpcodeTableByNumber[n];
        } else {
            return mOpcodeTableByNumber[OP_NOP];
        }
    }

    // If c is a valid opcode, returns the details for it, where the num field is the opcode number,
    // If c is not an opcode, returns the details for it where the num field is the char value
    //
    template <typename T>
    const typename machine<T>::opcodeDetails& machine<T>::findByShortMnemonic(char c) const
    {
        if ((unsigned)c >= 256) {
            // out of range of a char, return as if NOP
            return mOpcodeTableByNumber[OP_NOP];
        }

        return mOpcodeTableByCharacter[(unsigned)c];
    }

    /**
     * Executes the next Iota instruction. Depends on the current state of
     * Iota memory and the simulated registers @c PC, @c SP, and @c NZ.
     */
    template <typename T>
    bool machine<T>::executeOneCycle(void)
    {
        size_t temp;
        size_t Op0;
        size_t Op1;  // temporaries

        // Because the enumerator values are in a well ordered range in the
        // switch statement below, a reasonable compiler is able to generate
        // a jump table, reducing the switch statement overhead to a minimum.

        switch(mem[pc]) {
            default:
            case OP_NOP:
                incpc();
                break;

            case OP_RESET:
                pc = sp = 0;
                nz = false;
                break;

            case OP_HALT:
                return false;
                break;

            case OP_IN:
                if (mDataInIdx != mDataIn.size()) {
                    decsp();
                    setnz(mem[sp] = trunc((size_t)(T)mDataIn[mDataInIdx]));
                    // don't increment the index past the last element
                    mDataInIdx = std::min(mDataInIdx + 1, mDataIn.size() - 1);
                    incpc();
                } else if (mDataInIdx == 0 && mDataIn.size() == 0) {
                    // input zeros in lieu of an input stream
                    decsp();
                    setnz(mem[sp] = 0);
                    incpc();
                } else {
                    // HALT if reached end of input - should never happen
                    // because we prevent dataInIdx from exceeding the string length.
                    return false;
                }
                break;

            case OP_OUT:
                if (mOutputStream.size() < maxOutputStreamSize) {
                    mOutputStream += (char)(mem[sp] & 0xff);
                }

                if (mExecutionTraceLevel >= 1) {
                    std::cerr << "Output =====>(T) "
                         << std::to_string((unsigned char)(mem[sp] & 0xff))
                         << " = \'" << (char)(mem[sp] & 0xff) << "\'" << std::endl;
                }

                setnz(mem[sp] & 0xff);
                incsp();
                incpc();

                break;

            case OP_POP:
                setnz(mem[sp]);
                incsp();
                incpc();
                break;

            case OP_DUP:
                temp = mem[sp];
                decsp();
                setnz(mem[sp] = temp);
                incpc();
                break;

            case OP_PUSHPC:
                decsp();
                mem[sp] = trunc(pc);
                incpc();
                break;

            case OP_POPPC:
                pc = mem[sp] % mArchLength;
                incsp();
                break;

            case OP_POPSP:
                sp = mem[sp] % mArchLength;
                incpc();
                break;

            case OP_SPTGT: {
                // find the next TARGET
                bool found = false;
                if (pc < mArchLength - 1) {
                    for (auto i = pc + 1; !found && i < mArchLength; ++i) {
                        if (mem[i] == OP_TARGET) {
                            sp = i;
                            found = true;
                            break;
                        }
                    }
                }

                incpc();
                break; }

            case OP_PUSHNZ:
                decsp();
                mem[sp] = nz;
                incpc();
                break;

            case OP_SWAP:
                temp = mem[sp];
                mem[sp] = mem[(sp + 1) % mArchLength];
                mem[(sp + 1) % mArchLength] = temp;
                incpc();
                break;

            case OP_PUSH0:
                decsp();
                mem[sp] = 0;
                nz = 0;
                incpc();
                break;

            case OP_ADD:
                decsp();
                setnz(mem[sp] = trunc(mem[(sp + 2) % mArchLength] + mem[(sp + 1) % mArchLength]));
                incpc();
                break;

            case OP_SUB:
                decsp();
                setnz(mem[sp] = trunc(mem[(sp + 2) % mArchLength] - mem[(sp + 1) % mArchLength]));
                incpc();
                break;

            case OP_INC:
                mem[sp] = trunc(mem[sp] + 1);
                setnz(mem[sp]);
                incpc();
                break;

            case OP_DEC:
                mem[sp] = trunc(mem[sp] - 1);
                setnz(mem[sp]);
                incpc();
                break;

            case OP_MUL:
                decsp();
                setnz(mem[sp] = trunc(mem[(sp + 2) % mArchLength] * mem[(sp + 1) % mArchLength]));
                incpc();
                break;

            case OP_DIV: {
                Op0 = mem[(sp + 1) % mArchLength];
                Op1 = mem[sp];
                if (Op1 == 0) {
                    // in this case, fudge the operands to produce the largest result:
                    Op0 = ~0;
                    Op1 = 1;
                }
                setnz(mem[(sp + 1) % mArchLength] = trunc(Op0 / Op1));
                mem[sp] = trunc(Op0 % Op1);
                incpc();
                break; }

            case OP_XOR:
                decsp();
                setnz(mem[sp] = trunc(mem[(sp + 2) % mArchLength] ^ mem[(sp + 1) % mArchLength]));
                incpc();
                break;

            case OP_AND:
                decsp();
                setnz(mem[sp] = trunc(mem[(sp + 2) % mArchLength] & mem[(sp + 1) % mArchLength]));
                incpc();
                break;

            case OP_OR:
                decsp();
                setnz(mem[sp] = trunc(mem[(sp + 2) % mArchLength] | mem[(sp + 1) % mArchLength]));
                incpc();
                break;

            case OP_SHL:
                setnz(mem[sp] = trunc(mem[sp] << 1));
                incpc();
                break;

            case OP_SHR:
                setnz(mem[sp] = trunc(mem[sp] >> 1));
                incpc();
                break;

            case OP_NOT:
                setnz(mem[sp] = trunc(~mem[sp]));
                incpc();
                break;

            case OP_BZ:
                incpc();
                if (!nz) {
                    incpc();
                }
                break;

            case OP_BNZ:
                incpc();
                if (nz) {
                    incpc();
                }
                break;

            case OP_BEQ:
                incpc();
                if (mem[(sp + 1) % mArchLength] == mem[sp]) {
                    incpc();
                }
                break;

            case OP_BGT:
                incpc();
                if (mem[(sp + 1) % mArchLength] > mem[sp]) {
                    incpc();
                }
                break;

            case OP_BLT:
                incpc();
                if (mem[(sp + 1) % mArchLength] < mem[sp]) {
                    incpc();
                }
                break;

            case OP_BGE:
                incpc();
                if (mem[(sp + 1) % mArchLength] >= mem[sp]) {
                    incpc();
                }
                break;

            case OP_LOOP:
                // logic is in the std::endl opcode
                incpc();
                break;

            case OP_ENDL: {
                // find the preceding LOOP
                bool found = false;
                if (pc > 0) {
                    for (size_t i = pc; !found && i > 0; ) {
                        --i;
                        if (mem[i] == OP_LOOP) {
                            pc = (i + 1) % mArchLength;
                            found = true;
                            break;
                        }
                    }
                }
                if (!found) {
                    incpc();
                }
                break; }

            case OP_BRAN: {
                // find the next TARGET
                bool found = false;
                if (pc < mArchLength - 1) {
                    for (auto i = pc + 1; !found && i < mArchLength; ++i) {
                        if (mem[i] == OP_TARGET) {
                            pc = (i + 1) % mArchLength;
                            found = true;
                            break;
                        }
                    }
                }
                if (!found) {
                    incpc();
                }
                break; }

            case OP_BRAP: {
                // find the preceding TARGET
                bool found = false;
                if (pc > 0) {
                    for (size_t i = pc; !found && i > 0; ) {
                        --i;
                        if (mem[i] == OP_TARGET) {
                            pc = (i + 1) % mArchLength;
                            found = true;
                            break;
                        }
                    }
                }
                if (!found) {
                    incpc();
                }
                break; }

            case OP_TARGET:
                // logic is in the SPTGT, BRAN, and BRAP opcodes
                incpc();
                break;

            case OP_SKIP1: pc = (pc + 2) % mArchLength; break;
            case OP_SKIP2: pc = (pc + 3) % mArchLength; break;
            case OP_SKIP3: pc = (pc + 4) % mArchLength; break;
            case OP_SKIP4: pc = (pc + 5) % mArchLength; break;
            case OP_SKIP5: pc = (pc + 6) % mArchLength; break;
            case OP_SKIP6: pc = (pc + 7) % mArchLength; break;
            case OP_SKIP7: pc = (pc + 8) % mArchLength; break;
            case OP_SKIP8: pc = (pc + 9) % mArchLength; break;
            case OP_SKIP9: pc = (pc + 10) % mArchLength; break;
        };

        ++mCycleNumber;

        return true;
    }
} // end namespace iota

#endif  /* LITTLECOMPILER_H */
