/*
 * iota.cpp - command line driver for the Iota little language simulator
 *
 * Copyright 2012 David R. Miller
 *
 * This file is part of the Iota little language simulator.
 *
 * Iota is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Iota is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Iota.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact: http://www.millermattson.com/dave
 *
 * Compiles with gcc 4.7 with options:
 * -std=c++11 -W -Wall -Wextra -pedantic -D__GXX_EXPERIMENTAL_CXX0X__=1
 */

/*
 * See iota-machine.h for the interface to the Iota virtual machine.
 *
 * Command line synopsis:
 *     iota [-h|--help] [--version] [-L arch-length] [-W arch-width]
 *          [-k max-cycles] [-I input-data] [-S] [-T]
 *          [-d debug-level] program-source
 *
 *     -h
 *     --help
 *             Print the command line synopsis.
 *     --version
 *             Print the program version number.
 *     -L arch-length
 *             Number of memory locations in the virtual machine.
 *     -W arch-width
 *             Number of bits wide for the memory system.
 *     -k max-cycles
 *             Limit the number of Iota instructions executed.
 *     -I input-data
 *             Define a source of chars for the IN opcode.
 *     -S
 *             Print a disassembly of the Iota program currently in memory.
 *     -T
 *             Enable per-instruction trace of Iota program execution.
 *     -d debug-level
 *             Zero is no debug; nonzero causes extra output.
 *     program-source
 *             Iota program source.
 */

#include <iostream>
#include <cstdio>
#include <cstring>
#include <string>
#include "iota-machine.h"

using namespace iota;
using namespace std;

const string version = "Iota command line version 1.0";



// This struct holds all the command line options, provides a central
// place to define their defaults, and defines some functions for
// parsing the command line.
//
struct commandLineParameters
{
    commandLineParameters(void)
    		// Some defaults:
          : archLength(4),              // -L
            archWidth(8),               // -W
            maxCyclesToRun(200),        // -k
            inputData(string()),        // -I
            enableDisassembly(false),  // -S
            enableTrace(false),        // -T
            debugLevel(1),              // -d
            program(string(";H;;"))     // the only arg with no option letter
    { }

    void parseCommandLine(int argc, char **argv);
    void synopsisAndExit(void) {
        cerr << "iota [-h|--help] [--version] [-L arch-length] [-W arch-width] [-k max-cycles]" << endl;
        cerr << "[-I data-input-string] [-S] [-T] [-d debug-level] program-string" << endl;

        exit(1);
    }

    size_t archLength;
    size_t archWidth;
    size_t maxCyclesToRun;
    string inputData;
    bool   enableDisassembly;
    bool   enableTrace;
    int    debugLevel;
    string program;

private:
    size_t getSize_t(int &i, int argc, char **argv)
    {
        if (argv[i][2] != '\0') {
            return atol(&argv[i][2]);
        } else if (i < argc - 1) {
            return atol(argv[++i]);
        }
        synopsisAndExit();
        return 0;
    }

    int getInt(int &i, int argc, char **argv)
    {
        if (argv[i][2] != '\0') {
            return atoi(&argv[i][2]);
        } else if (i < argc - 1) {
            return atoi(argv[++i]);
        }
        synopsisAndExit();
        return 0;
    }

    string getString(int &i, int argc, char **argv)
    {
        if (argv[i][2] != '\0') {
            return string(&argv[i][2]);
        } else if (i < argc - 1) {
            return string(argv[++i]);
        }
        synopsisAndExit();
        return string("");
    }
};

// Sets all the values in the parameter block, or else issues
// a synopsis message and exits. If the same option appears multiple
// times, the last one takes precedence (and no warning is issued).
// No warning is issued for unrecognized options.
// The program source string is the only option without an option letter.
//
void commandLineParameters::parseCommandLine(int argc, char **argv)
{
	bool lengthOverride = false;

    for (int i = 1; i < argc; ++i) {
        if (std::strcmp(argv[i], "-h") == 0 || std::strcmp(argv[i], "--help") == 0) {
            synopsisAndExit();
        } else if (strcmp(argv[i], "--version") == 0) {
        	cout << iota::languageVersion << endl
        	     << iota::simulatorVersion << endl
        	     << version << endl;
        } else if (std::strncmp(argv[i], "-L", 2) == 0) {
            archLength = getSize_t(i, argc, argv);
            lengthOverride = true;
        } else if (std::strncmp(argv[i], "-W", 2) == 0) {
            archWidth = getSize_t(i, argc, argv);
        } else if (std::strncmp(argv[i], "-k", 2) == 0) {
            maxCyclesToRun = getSize_t(i, argc, argv);
        } else if (std::strncmp(argv[i], "-I", 2) == 0) {
            inputData = getString(i, argc, argv);
        } else if (std::strncmp(argv[i], "-S", 2) == 0) {
        	enableDisassembly = true;
        } else if (std::strncmp(argv[i], "-T", 2) == 0) {
            enableTrace = true;
        } else if (std::strncmp(argv[i], "-d", 2) == 0) {
            debugLevel = getInt(i, argc, argv);
        } else {
            program = string(argv[i]);
        }
    }

    // The archLength defaults to the program length if no -L option was specified:

    if (!lengthOverride) {
    	archLength = program.size();
    }
}


int main(int argc, char **argv)
{
    commandLineParameters args;
    args.parseCommandLine(argc, argv);

    cerr << "    archLength  = "   << to_string(args.archLength) << " locations" << endl;
    cerr << "     archWidth  = "   << to_string(args.archWidth) << " bits" << endl;
    cerr << "     maxCycles  = "   << to_string(args.maxCyclesToRun) << endl;
    cerr << "     inputData  = \"" << args.inputData << "\"" << endl;
    cerr << "  programSource = \"" << args.program << "\"" << endl;

    machine<unsigned short> myMachine(args.archLength, args.archWidth);

    myMachine.copyProgramToMachineMemory(args.program);
    if (args.enableDisassembly) {
        myMachine.disassemble();
    }

    myMachine.setDataInputSource(args.inputData);
    myMachine.setExecutionTrace(args.enableTrace ? 1 : 0);

    size_t cyclesExecuted = myMachine.executeAtMost(args.maxCyclesToRun);

    if (cyclesExecuted == args.maxCyclesToRun) {
        cerr << "Executed " << to_string(cyclesExecuted) << " cycles" << endl;
    } else {
        cerr << "Halted at cycle " << to_string(cyclesExecuted) << endl;
    }

    cerr << "Program output = \"" + myMachine.mOutputStream << "\"" << endl;
    cout << myMachine.mOutputStream << endl;

    return 0;
}
